@@ -189,7 +189,6 @@ static const MVMREPROps this_repr = {
     NULL, /* spesh */
     "VMIter", /* name */
     MVM_REPR_ID_MVMIter,
-    0, /* refs_frames */
     NULL, /* unmanaged_size */
     NULL, /* describe_refs */
 };
@@ -233,64 +232,68 @@ MVMObject * MVM_iter(MVMThreadContext *tc, MVMObject *target) {
                 MVMStaticFrame      *sf     = frame->static_info;
                 MVMLexicalRegistry **lexreg = sf->body.lexical_names_list;
                 MVMuint32 i;
-                for (i = 0; i < sf->body.num_lexicals; i++) {
-                    MVMuint32 idx  = lexreg[i]->value;
-                    MVMuint16 type = sf->body.lexical_types[idx];
-                    switch (type) {
-                        case MVM_reg_obj: {
-                            MVMObject *obj = frame->env[idx].o;
-                            if (!obj)
-                                obj = MVM_frame_vivify_lexical(tc, frame, idx);
-                            MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, obj);
-                            break;
+                MVMROOT(tc, frame, {
+                MVMROOT(tc, sf, {
+                    for (i = 0; i < sf->body.num_lexicals; i++) {
+                        MVMuint32 idx  = lexreg[i]->value;
+                        MVMuint16 type = sf->body.lexical_types[idx];
+                        switch (type) {
+                            case MVM_reg_obj: {
+                                MVMObject *obj = frame->env[idx].o;
+                                if (!obj)
+                                    obj = MVM_frame_vivify_lexical(tc, frame, idx);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, obj);
+                                break;
+                            }
+                            case MVM_reg_str: {
+                                MVMObject *bs = MVM_repr_box_str(tc, hll->str_box_type,
+                                    frame->env[idx].s);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bs);
+                                break;
+                            }
+                            case MVM_reg_int8: {
+                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
+                                    frame->env[idx].i8);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                                break;
+                            }
+                            case MVM_reg_int16: {
+                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
+                                    frame->env[idx].i16);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                                break;
+                            }
+                            case MVM_reg_int32: {
+                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
+                                    frame->env[idx].i32);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                                break;
+                            }
+                            case MVM_reg_int64: {
+                                MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
+                                    frame->env[idx].i64);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                                break;
+                            }
+                            case MVM_reg_num32: {
+                                MVMObject *bn = MVM_repr_box_num(tc, hll->num_box_type,
+                                    frame->env[idx].n32);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bn);
+                                break;
+                            }
+                            case MVM_reg_num64: {
+                                MVMObject *bn = MVM_repr_box_num(tc, hll->num_box_type,
+                                    frame->env[idx].n64);
+                                MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bn);
+                                break;
+                            }
+                            default:
+                                MVM_exception_throw_adhoc(tc,
+                                    "Unknown lexical type encountered while building context iterator");
                         }
-                        case MVM_reg_str: {
-                            MVMObject *bs = MVM_repr_box_str(tc, hll->str_box_type,
-                                frame->env[idx].s);
-                            MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bs);
-                            break;
-                        }
-                        case MVM_reg_int8: {
-                            MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                frame->env[idx].i8);
-                            MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                            break;
-                        }
-                        case MVM_reg_int16: {
-                            MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                frame->env[idx].i16);
-                            MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                            break;
-                        }
-                        case MVM_reg_int32: {
-                            MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                frame->env[idx].i32);
-                            MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                            break;
-                        }
-                        case MVM_reg_int64: {
-                            MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
-                                frame->env[idx].i64);
-                            MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
-                            break;
-                        }
-                        case MVM_reg_num32: {
-                            MVMObject *bn = MVM_repr_box_num(tc, hll->num_box_type,
-                                frame->env[idx].n32);
-                            MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bn);
-                            break;
-                        }
-                        case MVM_reg_num64: {
-                            MVMObject *bn = MVM_repr_box_num(tc, hll->num_box_type,
-                                frame->env[idx].n64);
-                            MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bn);
-                            break;
-                        }
-                        default:
-                            MVM_exception_throw_adhoc(tc,
-                                "Unknown lexical type encountered while building context iterator");
                     }
-                }
+                });
+                });
             });
 
             /* Call ourselves recursively to get the iterator for this
@@ -298,8 +301,8 @@ MVMObject * MVM_iter(MVMThreadContext *tc, MVMObject *target) {
             iterator = (MVMIter *)MVM_iter(tc, ctx_hash);
         }
         else {
-            MVM_exception_throw_adhoc(tc, "Cannot iterate object with %s representation",
-                REPR(target)->name);
+            MVM_exception_throw_adhoc(tc, "Cannot iterate object with %s representation (%s)",
+                REPR(target)->name, STABLE(target)->debug_name);
         }
     });
     return (MVMObject *)iterator;
@@ -324,7 +327,7 @@ MVMint64 MVM_iter_istrue(MVMThreadContext *tc, MVMIter *iter) {
 MVMString * MVM_iterkey_s(MVMThreadContext *tc, MVMIter *iterator) {
     if (REPR(iterator)->ID != MVM_REPR_ID_MVMIter
             || iterator->body.mode != MVM_ITER_MODE_HASH)
-        MVM_exception_throw_adhoc(tc, "This is not a hash iterator");
+        MVM_exception_throw_adhoc(tc, "This is not a hash iterator, it's a %s (%s)", REPR(iterator)->name, STABLE(iterator)->debug_name);
     if (!iterator->body.hash_state.curr)
         MVM_exception_throw_adhoc(tc, "You have not advanced to the first item of the hash iterator, or have gone past the end");
     return (MVMString *)iterator->body.hash_state.curr->key;
@@ -335,7 +338,7 @@ MVMObject * MVM_iterval(MVMThreadContext *tc, MVMIter *iterator) {
     MVMObject *target;
     MVMRegister result;
     if (REPR(iterator)->ID != MVM_REPR_ID_MVMIter)
-        MVM_exception_throw_adhoc(tc, "This is not an iterator");
+        MVM_exception_throw_adhoc(tc, "This is not an iterator, it's a %s (%s)", REPR(iterator)->name, STABLE(iterator)->debug_name);
     if (iterator->body.mode == MVM_ITER_MODE_ARRAY) {
         body = &iterator->body;
         if (body->array_state.index == -1)
