@@ -649,6 +649,7 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
                 sf->body.handlers[j].label_reg = read_int16(pos, 0);
                 pos += 2;
             }
+            sf->body.handlers[j].inlined_and_not_lexical = 0;
         }
     }
 
@@ -874,6 +875,9 @@ void MVM_bytecode_unpack(MVMThreadContext *tc, MVMCompUnit *cu) {
     cu_body->num_callsites = rs->expected_callsites;
     cu_body->orig_callsites = rs->expected_callsites;
 
+    if (rs->hll_str_idx > rs->expected_strings)
+        MVM_exception_throw_adhoc(tc, "Unpacking bytecode: HLL name string index out of range: %d > %d", rs->hll_str_idx, rs->expected_strings);
+
     /* Resolve HLL name. */
     MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->hll_name,
         MVM_cu_string(tc, cu, rs->hll_str_idx));
