@@ -313,24 +313,29 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
         MVMuint8 **backup_interp_bytecode_start = tc->interp_bytecode_start;
         MVMRegister **backup_interp_reg_base    = tc->interp_reg_base;
         MVMCompUnit **backup_interp_cu          = tc->interp_cu;
-        MVMFrame *backup_cur_frame              = tc->cur_frame;
+        MVMFrame *backup_cur_frame              = MVM_frame_force_to_heap(tc, tc->cur_frame);
         MVMFrame *backup_thread_entry_frame     = tc->thread_entry_frame;
-        MVMuint32 backup_mark                   = MVM_gc_root_temp_mark(tc);
-        jmp_buf backup_interp_jump;
-        memcpy(backup_interp_jump, tc->interp_jump, sizeof(jmp_buf));
+        MVMROOT(tc, backup_cur_frame, {
+        MVMROOT(tc, backup_thread_entry_frame, {
+            MVMuint32 backup_mark                   = MVM_gc_root_temp_mark(tc);
+            jmp_buf backup_interp_jump;
+            memcpy(backup_interp_jump, tc->interp_jump, sizeof(jmp_buf));
 
-        tc->cur_frame->return_value = &res;
-        tc->cur_frame->return_type  = MVM_RETURN_OBJ;
-        MVM_interp_run(tc, callback_invoke, &cid);
+            tc->cur_frame->return_value = &res;
+            tc->cur_frame->return_type  = MVM_RETURN_OBJ;
+            MVM_interp_run(tc, callback_invoke, &cid);
 
-        tc->interp_cur_op         = backup_interp_cur_op;
-        tc->interp_bytecode_start = backup_interp_bytecode_start;
-        tc->interp_reg_base       = backup_interp_reg_base;
-        tc->interp_cu             = backup_interp_cu;
-        tc->cur_frame             = backup_cur_frame;
-        tc->thread_entry_frame    = backup_thread_entry_frame;
-        memcpy(tc->interp_jump, backup_interp_jump, sizeof(jmp_buf));
-        MVM_gc_root_temp_mark_reset(tc, backup_mark);
+            tc->interp_cur_op         = backup_interp_cur_op;
+            tc->interp_bytecode_start = backup_interp_bytecode_start;
+            tc->interp_reg_base       = backup_interp_reg_base;
+            tc->interp_cu             = backup_interp_cu;
+            tc->cur_frame             = backup_cur_frame;
+            tc->current_frame_nr      = backup_cur_frame->sequence_nr;
+            tc->thread_entry_frame    = backup_thread_entry_frame;
+            memcpy(tc->interp_jump, backup_interp_jump, sizeof(jmp_buf));
+            MVM_gc_root_temp_mark_reset(tc, backup_mark);
+        });
+        });
     }
 
     /* Handle return value. */
