@@ -54,9 +54,6 @@ struct MVMThreadContext {
     /* Thread object representing the thread. */
     MVMThread *thread_obj;
 
-    /* The frame lying at the base of the current thread. */
-    MVMFrame *thread_entry_frame;
-
     /* Pointer to where the interpreter's current opcode is stored. */
     MVMuint8 **interp_cur_op;
 
@@ -71,9 +68,18 @@ struct MVMThreadContext {
      * is stored. */
     MVMCompUnit **interp_cu;
 
+    /* First call stack memory region, so we can traverse them for cleanup. */
+    MVMCallStackRegion *stack_first;
+
+    /* Current call stack region, which the next frame will be allocated in. */
+    MVMCallStackRegion *stack_current;
+
     /* The frame we're currently executing. */
     MVMFrame *cur_frame;
 
+    /* The frame lying at the base of the current thread. */
+    MVMFrame *thread_entry_frame;
+
     /* libuv event loop */
     uv_loop_t *loop;
 
@@ -89,6 +95,9 @@ struct MVMThreadContext {
     /* Result object of the last-run exception handler. */
     MVMObject *last_handler_result;
 
+    /* Last payload made available in a payload-goto exception handler. */
+    MVMObject *last_payload;
+
     /* Mutex that must be released if we throw an exception. Used in places
      * like I/O, which grab a mutex but may throw an exception. */
     uv_mutex_t *ex_release_mutex;
@@ -201,6 +210,11 @@ struct MVMThreadContext {
 
     /* Profiling data collected for this thread, if profiling is on. */
     MVMProfileThreadData *prof_data;
+
+    /* Frame sequence numbers in order to cheaply identify the place of a frame
+     * in the call stack */
+    MVMint32 current_frame_nr;
+    MVMint32 next_frame_nr;
 };
 
 MVMThreadContext * MVM_tc_create(MVMInstance *instance);
