@@ -23,7 +23,8 @@ MVMSpeshCandidate * MVM_spesh_candidate_setup(MVMThreadContext *tc,
     MVMint32 num_spesh_slots, num_log_slots, num_guards, *deopts, num_deopts;
     MVMuint16 num_locals, num_lexicals, used;
     MVMCollectable **spesh_slots, **log_slots;
-    char *before, *after;
+    char *before = 0;
+    char *after = 0;
     MVMSpeshGraph *sg;
 
     /* If we're profiling, log we're starting spesh work. */
@@ -32,7 +33,7 @@ MVMSpeshCandidate * MVM_spesh_candidate_setup(MVMThreadContext *tc,
 
     /* Do initial generation of the specialization, working out the argument
      * guards and adding logging. */
-    sg = MVM_spesh_graph_create(tc, static_frame, 0);
+    sg = MVM_spesh_graph_create(tc, static_frame, 0, 1);
     if (tc->instance->spesh_log_fh)
         before = MVM_spesh_dump(tc, sg);
     MVM_spesh_args(tc, sg, callsite, args);
@@ -110,14 +111,14 @@ MVMSpeshCandidate * MVM_spesh_candidate_setup(MVMThreadContext *tc,
                 fprintf(tc->instance->spesh_log_fh,
                     "Before:\n%s\nAfter:\n%s\n\n========\n\n", before, after);
                 fflush(tc->instance->spesh_log_fh);
-                MVM_free(before);
-                MVM_free(after);
                 MVM_free(c_name);
                 MVM_free(c_cuid);
             }
             used = 1;
         }
     }
+    MVM_free(after);
+    MVM_free(before);
     if (result && !used) {
         MVM_free(sc->bytecode);
         if (sc->handlers)
@@ -222,3 +223,20 @@ void MVM_spesh_candidate_specialize(MVMThreadContext *tc, MVMStaticFrame *static
     if (tc->instance->profiling)
         MVM_profiler_log_spesh_end(tc);
 }
+
+
+void MVM_spesh_candidate_destroy(MVMThreadContext *tc, MVMSpeshCandidate *candidate) {
+    if (candidate->sg)
+        MVM_spesh_graph_destroy(tc, candidate->sg);
+    MVM_free(candidate->guards);
+    MVM_free(candidate->bytecode);
+    MVM_free(candidate->handlers);
+    MVM_free(candidate->spesh_slots);
+    MVM_free(candidate->deopts);
+    MVM_free(candidate->log_slots);
+    MVM_free(candidate->inlines);
+    MVM_free(candidate->local_types);
+    MVM_free(candidate->lexical_types);
+    if (candidate->jitcode)
+        MVM_jit_destroy_code(tc, candidate->jitcode);
+}
