@@ -6,7 +6,7 @@
 
 /* Finds the index of a given codepoint within a trie node. Returns it if
  * there is one, or negative if there is not (note 0 is a valid index). */
-static MVMint32 find_child_node_idx(MVMThreadContext *tc, MVMNFGTrieNode *node, MVMCodepoint cp) {
+static MVMint32 find_child_node_idx(MVMThreadContext *tc, const MVMNFGTrieNode *node, MVMCodepoint cp) {
     if (node) {
         /* TODO: update this to do a binary search later on. */
         MVMint32 i;
@@ -18,7 +18,7 @@ static MVMint32 find_child_node_idx(MVMThreadContext *tc, MVMNFGTrieNode *node,
 }
 
 /* Does a lookup in the trie for a synthetic for the specified codepoints. */
-MVMNFGTrieNode * find_child_node(MVMThreadContext *tc, MVMNFGTrieNode *node, MVMCodepoint cp) {
+MVMNFGTrieNode * find_child_node(MVMThreadContext *tc, const MVMNFGTrieNode *node, MVMCodepoint cp) {
     MVMint32 idx = find_child_node_idx(tc, node, cp);
     return idx >= 0 ? node->next_codes[idx].node : NULL;
 }
@@ -144,7 +144,7 @@ static void add_synthetic_to_trie(MVMThreadContext *tc, MVMCodepoint *codes, MVM
  * checked that the synthetic does not exist. Adds it to the lookup trie and
  * synthetics table, making sure to do enough copy/free-at-safe-point work to
  * not upset other threads possibly doing concurrent reads. */
-static MVMGrapheme32 add_synthetic(MVMThreadContext *tc, MVMCodepoint *codes, MVMint32 num_codes) {
+static MVMGrapheme32 add_synthetic(MVMThreadContext *tc, MVMCodepoint *codes, MVMint32 num_codes, MVMint32 utf8_c8) {
     MVMNFGState     *nfg = tc->instance->nfg;
     MVMNFGSynthetic *synth;
     MVMGrapheme32    result;
@@ -153,9 +153,8 @@ static MVMGrapheme32 add_synthetic(MVMThreadContext *tc, MVMCodepoint *codes, MV
     /* Grow the synthetics table if needed. */
     if (nfg->num_synthetics % MVM_SYNTHETIC_GROW_ELEMS == 0) {
         size_t orig_size = nfg->num_synthetics * sizeof(MVMNFGSynthetic);
-        size_t new_size  = nfg->num_synthetics * sizeof(MVMNFGSynthetic);
-        MVMNFGSynthetic *new_synthetics = MVM_fixed_size_alloc(tc, tc->instance->fsa,
-            (nfg->num_synthetics + MVM_SYNTHETIC_GROW_ELEMS) * sizeof(MVMNFGSynthetic));
+        size_t new_size  = (nfg->num_synthetics + MVM_SYNTHETIC_GROW_ELEMS) * sizeof(MVMNFGSynthetic);
+        MVMNFGSynthetic *new_synthetics = MVM_fixed_size_alloc(tc, tc->instance->fsa, new_size);
         if (orig_size) {
             memcpy(new_synthetics, nfg->synthetics, orig_size);
             MVM_fixed_size_free_at_safepoint(tc, tc->instance->fsa, orig_size, nfg->synthetics);
@@ -170,10 +169,11 @@ static MVMGrapheme32 add_synthetic(MVMThreadContext *tc, MVMCodepoint *codes, MV
     comb_size        = synth->num_combs * sizeof(MVMCodepoint);
     synth->combs     = MVM_fixed_size_alloc(tc, tc->instance->fsa, comb_size);
     memcpy(synth->combs, codes + 1, comb_size);
-    synth->case_uc   = 0;
-    synth->case_lc   = 0;
-    synth->case_tc   = 0;
-    synth->case_fc   = 0;
+    synth->case_uc    = 0;
+    synth->case_lc    = 0;
+    synth->case_tc    = 0;
+    synth->case_fc    = 0;
+    synth->is_utf8_c8 = utf8_c8;
 
     /* Memory barrier to make sure the synthetic is fully in place before we
      * bump the count. */
@@ -193,13 +193,13 @@ static MVMGrapheme32 add_synthetic(MVMThreadContext *tc, MVMCodepoint *codes, MV
 /* Does a lookup of a synthetic in the trie. If we find one, returns it. If
  * not, acquires the update lock, re-checks that we really are missing the
  * synthetic, and then adds it. */
-static MVMGrapheme32 lookup_or_add_synthetic(MVMThreadContext *tc, MVMCodepoint *codes, MVMint32 num_codes) {
+static MVMGrapheme32 lookup_or_add_synthetic(MVMThreadContext *tc, MVMCodepoint *codes, MVMint32 num_codes, MVMint32 utf8_c8) {
     MVMGrapheme32 result = lookup_synthetic(tc, codes, num_codes);
     if (!result) {
         uv_mutex_lock(&tc->instance->nfg->update_mutex);
         result = lookup_synthetic(tc, codes, num_codes);
         if (!result)
-            result = add_synthetic(tc, codes, num_codes);
+            result = add_synthetic(tc, codes, num_codes, utf8_c8);
         uv_mutex_unlock(&tc->instance->nfg->update_mutex);
     }
     return result;
@@ -214,7 +214,22 @@ MVMGrapheme32 MVM_nfg_codes_to_grapheme(MVMThreadContext *tc, MVMCodepoint *code
     if (num_codes == 1)
         return codes[0];
     else
-        return lookup_or_add_synthetic(tc, codes, num_codes);
+        return lookup_or_add_synthetic(tc, codes, num_codes, 0);
+}
+
+/* Does the same as MVM_nfg_codes_to_grapheme, but flags the added grapheme as
+ * being an UTF8-C8 synthetic. */
+MVMGrapheme32 MVM_nfg_codes_to_grapheme_utf8_c8(MVMThreadContext *tc, MVMCodepoint *codes, MVMint32 num_codes) {
+    if (num_codes == 1)
+        return codes[0];
+    else
+        return lookup_or_add_synthetic(tc, codes, num_codes, 1);
+}
+
+/* Gets the \r\n synthetic. */
+MVMGrapheme32 MVM_nfg_crlf_grapheme(MVMThreadContext *tc) {
+    MVMCodepoint codes[2] = { '\r', '\n' };
+    return lookup_or_add_synthetic(tc, codes, 2, 0);
 }
 
 /* Does a lookup of information held about a synthetic. The synth parameter
@@ -233,49 +248,103 @@ MVMNFGSynthetic * MVM_nfg_get_synthetic_info(MVMThreadContext *tc, MVMGrapheme32
 
 /* Gets the cached case change if we already computed it, or computes it if
  * this is the first time we're using it. */
-static MVMGrapheme32 compute_case_change(MVMThreadContext *tc, MVMGrapheme32 synth, MVMNFGSynthetic *synth_info, MVMint32 case_) {
-    MVMCodepoint  lowered_base = MVM_unicode_get_case_change(tc, synth_info->base, case_);
-    MVMGrapheme32 result;
-    if (lowered_base == synth_info->base) {
-        /* Base character does not change, so grapheme stays the same. */
-        result = synth;
+static MVMGrapheme32 CASE_UNCHANGED[1] = {0};
+static void compute_case_change(MVMThreadContext *tc, MVMGrapheme32 synth, MVMNFGSynthetic *synth_info, MVMint32 case_) {
+    MVMGrapheme32 *result;
+    MVMint32 num_result_graphs;
+
+    /* Transform the base character. */
+    const MVMCodepoint *result_cps;
+    MVMuint32     num_result_cps = MVM_unicode_get_case_change(tc, synth_info->base,
+        case_, &result_cps);
+    if (num_result_cps == 0 || *result_cps == synth_info->base) {
+        /* Base character does not change, so grapheme stays the same. We
+         * install a non-null sentinel for this case, and set the result
+         * grapheme count to zero, which indicates no change. */
+        result = CASE_UNCHANGED;
+        num_result_graphs = 0;
     }
     else {
-        /* Build a codepoint buffer with the lowered base and then either
-         * lookup or create a synthetic as needed. */
-        MVMint32      num_codes     = 1 + synth_info->num_combs;
-        MVMCodepoint *change_buffer = MVM_malloc(num_codes * sizeof(MVMCodepoint));
-        change_buffer[0] = lowered_base;
-        memcpy(change_buffer + 1, synth_info->combs, synth_info->num_combs * sizeof(MVMCodepoint));
-        result = MVM_nfg_codes_to_grapheme(tc, change_buffer, num_codes);
-        MVM_free(change_buffer);
+        /* We can potentially get multiple graphemes back. We may also get
+         * into situations where we case change the base and suddenly we
+         * can normalize the whole thing to a non-synthetic. So, we take
+         * a trip through the normalizer. Note we push the first thing
+         * we get back from the case change, then our combiners, and
+         * finally anything else the case change produced. This should
+         * do about the right thing for both case changes that produce a
+         * base and a combiner, and those that produce a base and a base,
+         * since the normalizer applies Unicode canonical sorting. */
+        MVMNormalizer norm;
+        MVMint32 i;
+        MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFG);
+        MVM_unicode_normalizer_push_codepoints(tc, &norm, result_cps, 1);
+        MVM_unicode_normalizer_push_codepoints(tc, &norm, synth_info->combs,
+            synth_info->num_combs);
+        if (num_result_cps > 1)
+            MVM_unicode_normalizer_push_codepoints(tc, &norm, result_cps + 1,
+                num_result_cps - 1);
+        MVM_unicode_normalizer_eof(tc, &norm);
+
+        num_result_graphs = MVM_unicode_normalizer_available(tc, &norm);
+        result = MVM_malloc(num_result_graphs * sizeof(MVMGrapheme32));
+        for (i = 0; i < num_result_graphs; i++)
+            result[i] = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+        MVM_unicode_normalizer_cleanup(tc, &norm);
+    }
+
+    switch (case_) {
+    case MVM_unicode_case_change_type_upper:
+        synth_info->case_uc = result;
+        synth_info->case_uc_graphs = num_result_graphs;
+        break;
+    case MVM_unicode_case_change_type_lower:
+        synth_info->case_lc = result;
+        synth_info->case_lc_graphs = num_result_graphs;
+        break;
+    case MVM_unicode_case_change_type_title:
+        synth_info->case_tc = result;
+        synth_info->case_tc_graphs = num_result_graphs;
+        break;
+    case MVM_unicode_case_change_type_fold:
+        synth_info->case_fc = result;
+        synth_info->case_fc_graphs = num_result_graphs;
+        break;
+    default:
+        MVM_panic(1, "NFG: invalid case change %d", case_);
     }
-    return result;
 }
-MVMGrapheme32 MVM_nfg_get_case_change(MVMThreadContext *tc, MVMGrapheme32 synth, MVMint32 case_) {
+MVMuint32 MVM_nfg_get_case_change(MVMThreadContext *tc, MVMGrapheme32 synth, MVMint32 case_, MVMGrapheme32 **result) {
     MVMNFGSynthetic *synth_info = MVM_nfg_get_synthetic_info(tc, synth);
     switch (case_) {
     case MVM_unicode_case_change_type_upper:
         if (!synth_info->case_uc)
-            synth_info->case_uc = compute_case_change(tc, synth, synth_info, case_);
-        return synth_info->case_uc;
+            compute_case_change(tc, synth, synth_info, case_);
+        *result = synth_info->case_uc;
+        return synth_info->case_uc_graphs;
     case MVM_unicode_case_change_type_lower:
         if (!synth_info->case_lc)
-            synth_info->case_lc = compute_case_change(tc, synth, synth_info, case_);
-        return synth_info->case_lc;
+            compute_case_change(tc, synth, synth_info, case_);
+        *result = synth_info->case_lc;
+        return synth_info->case_lc_graphs;
     case MVM_unicode_case_change_type_title:
         if (!synth_info->case_tc)
-            synth_info->case_tc = compute_case_change(tc, synth, synth_info, case_);
-        return synth_info->case_tc;
+            compute_case_change(tc, synth, synth_info, case_);
+        *result = synth_info->case_tc;
+        return synth_info->case_tc_graphs;
+    case MVM_unicode_case_change_type_fold:
+        if (!synth_info->case_fc)
+            compute_case_change(tc, synth, synth_info, case_);
+        *result = synth_info->case_fc;
+        return synth_info->case_fc_graphs;
     default:
-        abort();
+        MVM_panic(1, "NFG: invalid case change %d", case_);
     }
 }
 
 /* Returns non-zero if the result of concatenating the two strings will freely
  * leave us in NFG without any further effort. */
 static MVMint32 passes_quickcheck_and_zero_ccc(MVMThreadContext *tc, MVMCodepoint cp) {
-    const char *qc_str  = MVM_unicode_codepoint_get_property_cstr(tc, cp, MVM_UNICODE_PROPERTY_NFC_QC);
+    const char *qc_str  = MVM_unicode_codepoint_get_property_cstr(tc, cp, MVM_UNICODE_PROPERTY_NFG_QC);
     const char *ccc_str = MVM_unicode_codepoint_get_property_cstr(tc, cp, MVM_UNICODE_PROPERTY_CANONICAL_COMBINING_CLASS);
     return qc_str && qc_str[0] == 'Y' &&
         (!ccc_str || strlen(ccc_str) > 3 || (strlen(ccc_str) == 1 && ccc_str[0] == 0));
@@ -298,10 +367,47 @@ MVMint32 MVM_nfg_is_concat_stable(MVMThreadContext *tc, MVMString *a, MVMString
     if (last_a < 0 || first_b < 0)
         return 0;
 
-    /* If both less than the first significant char for NFC, we're good. */
-    if (last_a < MVM_NORMALIZE_FIRST_SIG_NFC && first_b < MVM_NORMALIZE_FIRST_SIG_NFC)
+    /* If both less than the first significant char for NFC, and the first is
+     * not \r, we're good. */
+    if (last_a != 0x0D && last_a < MVM_NORMALIZE_FIRST_SIG_NFC
+                       && first_b < MVM_NORMALIZE_FIRST_SIG_NFC)
         return 1;
 
     /* If either fail quickcheck or have ccc > 0, have to re-normalize. */
     return passes_quickcheck_and_zero_ccc(tc, last_a) && passes_quickcheck_and_zero_ccc(tc, first_b);
 }
+
+/* Free all memory allocated to hold synthetic graphemes. These are global
+ * to a VM instance. */
+void MVM_nfg_destroy(MVMThreadContext *tc) {
+    MVMNFGState *nfg = tc->instance->nfg;
+    MVMint32 i;
+
+    /* Free all synthetics. */
+    if (nfg->synthetics) {
+        size_t used_synths_in_block = nfg->num_synthetics % MVM_SYNTHETIC_GROW_ELEMS;
+        size_t synths_to_free = used_synths_in_block
+            ? nfg->num_synthetics + (MVM_SYNTHETIC_GROW_ELEMS - used_synths_in_block)
+            : nfg->num_synthetics;
+
+        for (i = 0; i < nfg->num_synthetics; i++) {
+            MVM_fixed_size_free(tc, tc->instance->fsa,
+                nfg->synthetics[i].num_combs * sizeof(MVMCodepoint),
+                nfg->synthetics[i].combs);
+            if (nfg->synthetics[i].case_uc != CASE_UNCHANGED)
+                MVM_free(nfg->synthetics[i].case_uc);
+            if (nfg->synthetics[i].case_lc != CASE_UNCHANGED)
+                    MVM_free(nfg->synthetics[i].case_lc);
+            if (nfg->synthetics[i].case_tc != CASE_UNCHANGED)
+                MVM_free(nfg->synthetics[i].case_tc);
+            if (nfg->synthetics[i].case_fc != CASE_UNCHANGED)
+                MVM_free(nfg->synthetics[i].case_fc);
+        }
+
+        MVM_fixed_size_free(tc, tc->instance->fsa,
+            synths_to_free * sizeof(MVMNFGSynthetic),
+            nfg->synthetics);
+    }
+
+    MVM_free(nfg);
+}
