@@ -17,6 +17,7 @@ typedef struct MVMCallCapture MVMCallCapture;
 typedef struct MVMCallCaptureBody MVMCallCaptureBody;
 typedef struct MVMCallsite MVMCallsite;
 typedef struct MVMCallsiteInterns MVMCallsiteInterns;
+typedef struct MVMCallStackRegion MVMCallStackRegion;
 typedef struct MVMCFunction MVMCFunction;
 typedef struct MVMCFunctionBody MVMCFunctionBody;
 typedef struct MVMCode MVMCode;
@@ -104,7 +105,7 @@ typedef struct MVMCUnionNameMap MVMCUnionNameMap;
 typedef struct MVMCUnionREPRData MVMCUnionREPRData;
 typedef struct MVMMultiCache MVMMultiCache;
 typedef struct MVMMultiCacheBody MVMMultiCacheBody;
-typedef struct MVMMultiArityCache MVMMultiArityCache;
+typedef struct MVMMultiCacheNode MVMMultiCacheNode;
 typedef struct MVMMultiDimArray MVMMultiDimArray;
 typedef struct MVMMultiDimArrayBody MVMMultiDimArrayBody;
 typedef struct MVMMultiDimArrayREPRData MVMMultiDimArrayREPRData;
